/*  serial port interface for Palm Portable Keyboard
 *  Copyright (C) 2003 
 *  Trap <CXL00145@nifty.com> (linux zaurus SL-C7x0 version)
 *
 *  Based on ser_code.c created by ...
 *    Dan Morris <dmorris@cs.brown.edu> (initial and windows version)
 *    Ralf Ackermann <rac@KOM.tu-darmstadt.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  2003.12.14 v1.1.0 : Performance Up.
 *    The weight of the value of the parameter (keyrep_first, keyrep_next)
 *    was changed. When using -d option, a default config file changed into
 *    ser_code_direct.ini automatically.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <libgen.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <termio.h>
#include <error.h>

// mouse
#include <linux/fb.h>
#include <asm/page.h>	
#include <sys/mman.h>
// end_mouse

#ifndef errno
extern int errno;
#endif

#include <asm/sharp_char.h>
#include <asm/sharp_keycode.h>
#include <linux/kbd_ll.h>



/*
keyboard protocol is:
   key press   => key scan code
   key release => (key scan code | (1 << 7)) (most significant bit set)

   When all key was released, last release code once again arrived.
*/


// Constants

#define	MYFLG_SHIFT	0x01
#define	MYFLG_CTRL	0x02
#define	MYFLG_FN	0x04
#define	MYFLG_CMD	0x08

#define RETOUR_OK	0
#define RETOUR_ERROR	-1

#define	RTS_TIMEOUT	600

#define VERSION		"1.1.0"

#define	VNCKBDDRV	"/dev/kbdsim"
#define	SHARPKBDDRV	"/dev/sharp_kbdctl"

// user structure

typedef struct key_mapping {
    char	*keyname;
    long	keycode;
} key_mappings;

// global variables

static int		fdd = -1;		// IOCTL ԤǥХѤΥեϥɥ
static int		fds = -1;		// ܡɤμѥեϥɥ
static int		directflag = 0;		// 쥯ȥ⡼ɤΥե饰
static int		infoflag = 0;		// debug åɽե饰
static int		vncflag = 0;		// VNCɥ饤ѥ⡼ɤΥե饰
static int		keyrepeat = 0;		// ԡȤΤΥ
static int		keyrep_first = 3;	// 󥭡ԡԤ(0.3s)
static int		keyrep_next = 1;	// 2ܰʹߥԡԤ(0.1s)
static long		lastkeycode = -1;	// Ǹ˲줿Ф
static char *		program_name = "";	// Υץ̾

static char *		serialdevice = "/dev/ttyS0";
static char *		keymapfile   = NULL;
static char *		keymapfile_n = "/opt/QtPalmtop/etc/ser_code_ksim.ini";
static char *		keymapfile_d = "/opt/QtPalmtop/etc/ser_code_direct.ini";
static char *		metas_str[8] = { "lshift", "lctrl", "lfn", "lcmd", "rshift", "rctrl", "rfn", "rcmd" };

// Ǽΰ
static int		metaflags;		// meta-key β֤򵭲ѿ
static char		keys[128];		// ƥ ON/OFF ֤ݻ
static unsigned char	metas[128];		// Υɤ meta-key Ǥ뤫򼨤
static key_mappings	keymap[16][128];	// cmd, fn, ctrl, shift ȹ礻16̤ 128 ĤΥ

// mouse
static int mousemode = 0;
static int xmode = 0;

static FILE *fp;

static int console_fd;
static struct fb_var_screeninfo vinfo;
static struct fb_fix_screeninfo finfo;
static int mapped_memlen;
static int mapped_offset;
static short *mapped_mem;
static short *mapped_mem_start;
static short *mapped_mem_end;

    int fb_offset;
//end_mouse


// prototypes

void		infoarea_init( void );
int		check_meta( const char *str );
int		read_key_mapping( char *map_file );

void		rts_down_up( int fd );
int		init_ser( char *ser_device );
int		set_ser_param( int fd );
void		flush_ser( int fd );

unsigned char	get_1_char( void );

void		sig_handler( int status );

void		close_all( void );
void		usage( int status );
void		printversion( void );
void		printinfo( void );

void		onekey_v( int c, int fd, int pressflag );
void		onekey_s( long c, int fd );
int		onekey( long c, int fdd );


//  mouse

#define POINTERCAL "/etc/pointercal"

static struct qt_cal {
  int ok;
  int a, b, c, d, e, f, s;
} cal;

static struct q_cal {
  int ok;
  double a, b, c, d, e, f, s;
} qcal;

static short pointer[16][16] = { 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
				 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
				 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
				 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
				 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
				 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
				 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
				 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
				 0x0000, 0x0000, 0x0000, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001,
				 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001,
				 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001,
				 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001,
				 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001,
				 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001,
				 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001,
				 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001 };

static short prev_image1[16][16];
static short prev_image2[16][16];
static short prev_x,prev_y;
static int st=0;

void mouseon(int x,int y)
{
  int i,j;
  
  x -= 16;
  for (i=0 ; i<16; i++){
    for (j=0 ; j<16; j++){
      if((mapped_mem+i+x+(j+y)*finfo.line_length/2) > mapped_mem_start &&
	 (mapped_mem+i+x+(j+y)*finfo.line_length/2) < mapped_mem_end &&
         ( (i + x) < 480 ) && ( (i + x) >= 0 ) && ( (j + y) < 640 ) && ( (j + y) >= 0 )  ){
	if (pointer[15-i][j] != 1) {
	  *(mapped_mem+i+x+(j+y)*finfo.line_length/2 ) = pointer[15-i][j];
	}
      }
    }
  }
  
return;
}

void mouseoff(int x,int y)
{
  int i,j;
  
  x -= 16;
  for (i=0 ; i<16; i++){
    for (j=0 ; j<16; j++){
      if((mapped_mem+i+x+(j+y)*finfo.line_length/2) > mapped_mem_start &&
	 (mapped_mem+i+x+(j+y)*finfo.line_length/2) < mapped_mem_end && 
         ( (i + x) < 480 ) && ( (i + x) >= 0 ) && ( (j + y) < 640 ) && ( (j + y) >= 0 ) &&
	*(mapped_mem+i+x+(j+y)*finfo.line_length/2 ) == prev_image2[i][j] ) {  
	  *(mapped_mem+i+x+(j+y)*finfo.line_length/2 ) = prev_image1[i][j];
      }
    }
  }
}

void grabrect1(int x,int y)
{
  int i,j;
  
  x -= 16;
  for (i=0 ; i<16; i++){
    for (j=0 ; j<16; j++){
      if((mapped_mem+i+x+(j+y)*finfo.line_length/2) > mapped_mem_start &&
	 (mapped_mem+i+x+(j+y)*finfo.line_length/2) < mapped_mem_end &&
         ( (i + x) < 480 ) && ( (i + x) >= 0 ) && ( (j + y) < 640 ) && ( (j + y) >= 0 )  ){
	prev_image1[i][j]=*(mapped_mem+i+x+(j+y)*finfo.line_length/2 );
      }
    }
  }
  
}

void grabrect2(int x,int y)
{
  int i,j;
  
  x -= 16;
  for (i=0 ; i<16; i++){
    for (j=0 ; j<16; j++){
      if((mapped_mem+i+x+(j+y)*finfo.line_length/2) > mapped_mem_start &&
	 (mapped_mem+i+x+(j+y)*finfo.line_length/2) < mapped_mem_end &&
         ( (i + x) < 480 ) && ( (i + x) >= 0 ) && ( (j + y) < 640 ) && ( (j + y) >= 0 )  ){
	prev_image2[i][j]=*(mapped_mem+i+x+(j+y)*finfo.line_length/2 );
      }
    }
  }
  
}

void loadConfig()
{
  FILE *pcal;
  cal.ok = 0;
  pcal = fopen(POINTERCAL, "r");
  if(pcal == NULL){
    fprintf(stderr, "Warning: cannot open " POINTERCAL ".\n");
  }else{
    int n;
    n = fscanf(pcal, "%d %d %d %d %d %d %d",
	       &cal.a, &cal.b, &cal.c, &cal.d, &cal.e, &cal.f, &cal.s);
    qcal.a = cal.a /1000;
    qcal.b = cal.b /1000;
    qcal.c = cal.c /1000;
    qcal.d = cal.d /1000;
    qcal.e = cal.e /1000;
    qcal.f = cal.f /1000;
    qcal.s = cal.s /1000;
    if(n != 7){
      fprintf(stderr, "Warning: " POINTERCAL " is unknown format.\n");
    }else{
      cal.ok = 1;
    }
    fclose(pcal);
  }
  
  
  console_fd = open("/dev/fb0", O_RDWR, 0);
  if ( console_fd < 0 ) {
    return;
  }
  
  if ( ioctl(console_fd, FBIOGET_FSCREENINFO, &finfo) < 0 ) {
    printf("error");
    return;
  }
  
  if ( ioctl(console_fd, FBIOGET_VSCREENINFO, &vinfo) < 0 ) {
    printf("error");
    return;
  }
  
  mapped_offset = (((long)finfo.smem_start) -
		   (((long)finfo.smem_start)&~(PAGE_SIZE-1)));
  mapped_memlen = finfo.smem_len+mapped_offset;
  mapped_mem = (short*)mmap(NULL, mapped_memlen,PROT_READ|PROT_WRITE, MAP_SHARED, console_fd, 0);
  mapped_mem_start = mapped_mem;
  mapped_mem_end = mapped_mem + finfo.smem_len;
  
  
}

// mouse


/*
  Ǽΰ̥륯ꥢ
*/
void	infoarea_init( void )
{
  metaflags = 0;
  memset( keys, 0, sizeof(keys) );
  memset( metas, 0, sizeof(metas) );
  memset( keymap, 0, sizeof(keymap) );
}


/*
  Ϳ줿ʸ󤬡ɤ meta-key ˰פ뤫Ĵ١
  metaflags ͤ(bitñ̤Υե饰)
  
  ret	== 0 : meta-key ǤϤʤ
  != 0 : meta-key Ǥ롣ͤ bitñ̤Υե饰͡
*/
int	check_meta( const char *str )
{
  int	i;
  int	flags = 0;
  
  for( i = 0; i < sizeof(metas_str)/sizeof(metas_str[0]); i++ ) {
    if ( strcasecmp( str, metas_str[i] ) == 0 ) {
      flags = 1 << i;
    }
  }
  
  return( flags );
}


/*
  ե뤫顢ޥåԥ󥰤ɤ߹
  
  ret	RETOUR_OK	ｪλ(ޥåץեȤϤȤ⤫)
  RETOUR_ERROR	۾ｪλ(ޥåץե뤬ʤ)
*/
int	read_key_mapping( char *map_file )
{
  int		flags;
  int		rcnt;
  unsigned int	id;
  long		keycode;
  FILE		*in_file;
  char		line[80];
  char		keyname[80];
  char		flagstr[80];
  
  
  // ѿꥢ
  infoarea_init();
  memset( line, 0, sizeof(line) );
  
  // ե򳫤
  if ( NULL == (in_file = fopen( map_file, "r" )) ) {
    return( RETOUR_ERROR );
  }
  
  // ե1Ԥɤߤʤ
  while ( ! feof(in_file) ) {
    
    // 1ɤ߹ (line byteʾԵ)
    if ( NULL != fgets(line, sizeof(line)-1, in_file) ) {
      
      // Ƭ # ιԤϥȹԤȤ̵뤹롣
      if ( line[0] == '#' ) continue;
      
      // meta-key ԤΤĤɤ߹Ǥߤ
      rcnt = sscanf(line, "%s %lx", keyname, &id );
      
      // Ƭιܤʤ̵뤷ƼιԤ(Ԥʤ)
      if ( rcnt < 1 ) continue;
      
      // ιԤ meta-key ԤǤСɤ߼ä metas ꤹ롣
      if ( (flags = check_meta( keyname )) != 0 ) {
	if ( (rcnt == 2) && (id < 128) ) metas[ id ] = flags;
	continue;
      }
      
      // ¾Υѥ᡼
      if ( strcasecmp( keyname, "keyrep_first" ) == 0 ) {
	keyrep_first = id & 0x7f;
	continue;
      }
      
      if ( strcasecmp( keyname, "keyrep_next" ) == 0 ) {
	keyrep_next = id & 0x7f;
	continue;
      }
      
      // Ƭܤʸ 16ʿʸǤʤȴ롣
      if ( (strlen(keyname) != 2) || (! isxdigit(keyname[0])) || (! isxdigit(keyname[1])) ) {
	continue; 
      }
      
      // ԤȤƲեޥåȤɤ߼롢
      rcnt = sscanf( line, "%x %s %s %lx", &id, keyname, flagstr, &keycode );
      
      // 4ɤʤФ
      if ( rcnt < 4 ) continue;
      
      // ϰϥå
      if ( id > 127 ) continue;
      
      // ե饰 S, C, F, M, N Σʸȹ礻롣 ex. SCF
      flags = 0;
      if ( strchr( flagstr, 'S' ) != NULL ) flags |= MYFLG_SHIFT;
      if ( strchr( flagstr, 'C' ) != NULL ) flags |= MYFLG_CTRL;
      if ( strchr( flagstr, 'F' ) != NULL ) flags |= MYFLG_FN;
      if ( strchr( flagstr, 'M' ) != NULL ) flags |= MYFLG_CMD;
      
      keymap[flags][id].keyname	= strdup(keyname);
      keymap[flags][id].keycode	= keycode;
    }
  }
  
  fclose( in_file );
  
  return( RETOUR_OK );
}


/*
  RTS 饤򲼤ơ夲롣
  (palm stowaway keyboard Υ꡼פ뤿ν)
*/
void	rts_down_up( int fd )
{
  int line;
  
  ioctl (fd, TIOCMGET, &line);
  line &= ~TIOCM_RTS;
  ioctl (fd, TIOCMSET, &line);
  
  if ( infoflag ) printf( "RTS set to LOW\n" );
  
  usleep(1000);
  
  ioctl (fd, TIOCMGET, &line);
  line |= TIOCM_RTS;
  ioctl (fd, TIOCMSET, &line);
  
  if ( infoflag ) printf( "RTS set to HIGH\n" );
}


/*
  ꤵ줿ꥢݡȤ¾ץ
  
  in	char	*ser_device	ꥢǥХ
  
  ret	RETOUR_OK	
  RETOUR_ERROR	ץ or ¾å or ѥ᡼꼺
  
  Хѿ fdd  -1 ʳǤХץ󤵤Ƥ롣
   directflag 򻲾Ȥƥ⡼ɤѤƤ롣
*/
int	init_ser( char *ser_device )
{
  // ser_device Υץ
  if ( (fds = open(ser_device, O_RDWR | O_NOCTTY, 0)) == -1 ) {
    printf( "Error opening serial device !\n" );
    return( RETOUR_ERROR );
  }
  
  // ¾ˤĤޤʤ褦¾å
  if ( flock( fds, LOCK_EX | LOCK_NB ) == -1 ) {
    return( RETOUR_ERROR );
  }
  
  // ѥ᡼
  if ( set_ser_param(fds) == RETOUR_ERROR ) {
    printf( "Couldn't set serial parameters !\n" );
    return( RETOUR_ERROR );
  }else{
    return( RETOUR_OK );
  }
}


/*
  ꥢݡȤΥԡɤ¾ (9600bps )
	 directflag 򻲾Ȥƥ⡼ɤѤƤ롣
*/
int	set_ser_param( int fd )
{
  int		line;
  struct termio	sgtty;

  if (ioctl(fd, TCGETA, &sgtty) < 0) {
    printf( "can't TCGETA\n" );
    return( RETOUR_ERROR );
  }
  
  sgtty.c_iflag		=  (IMAXBEL | IGNBRK);
  sgtty.c_oflag		=  0;
  sgtty.c_lflag		=  0;
  sgtty.c_cflag		=  (CLOCAL | HUPCL | B9600 | CS8 | CREAD);
  sgtty.c_cflag		&= ~(PARENB | PARODD);
  sgtty.c_cc[VTIME]	=  1;		/* ॢȤ 0.1 s	*/
  sgtty.c_cc[VMIN]	=  0;		/* ʸޤǥ֥å */
  
  if ( ioctl(fd, TCSETA, &sgtty) < 0 ) {
    printf( "can't TCSETA\n" );
    return( RETOUR_ERROR );
  }
  
  ioctl( fd, TIOCMGET, &line );
  line |= TIOCM_DTR;
  ioctl( fd, TIOCMSET, &line );
  
  if ( infoflag ) printf( "DTR set to HIGH\n" );
  
  sleep( 1 );
  rts_down_up( fd );	// stowaway keyboard enable
  flush_ser( fd );
  
  return( RETOUR_OK );
}


/*
  ϥɥΥեå
*/
void	flush_ser( int fd )
{
  ioctl( fd, TCFLSH, 0 );
}


/*
  1ʸϴؿ
  
  ǥȥԡȽԤäƤ롣
*/
unsigned char	get_1_char( void )
{
  unsigned char	recv_char;
  int		r_code;
  static int	timeout = RTS_TIMEOUT;
  
  for(;;) {
    /* wait for a char to arrive */
    r_code = read( fds, &recv_char, 1 );
    
    if ( r_code > 0 ) break;		// 줿/Υ줿ϥ롼פȴ
    
    // ȴƤ硢0.1 s вᤷƤ롣
    
    // ԡȽ (̾⡼ɻΤͭ)
    if ( lastkeycode > 0 ) {	// Ǹ˲줿ͭǡ
      keyrepeat --;			// ԡԤۤ饭ɤ
      
      if ( keyrepeat <= 0 ) {
	onekey( lastkeycode, fdd );
	keyrepeat = keyrep_next;
      }
    }
    
    // ⤵ʤΡRTS 夲 (1ʬ֤줿)
    timeout --;
    if ( timeout <= 0 ) {
      rts_down_up( fds );	// stowaway keyboard enable
      timeout = RTS_TIMEOUT;
    }
  }
  
  return( recv_char );
}


// Control-C ʥϥɥ
void	sig_handler( int status )
{
  printf( "SIGNAL(%d) ΤǥݡȤĤƽλޤ\n", status );
  exit( 1 );
}


/*
  ץλθ
*/
void	close_all( void )
{
  int	i;
  long	keycode;
  
  // open ѤΥϥɥĤƽλ
  
  // ꥢݡ
  if ( fds != -1 ) {
    close( fds );
    fds = -1;
  }
  
  // ǥХɥ饤
  if ( fdd != -1 ) {
    // ǰΤˡƤΥΥ
    
    for( i = 0; i < 127; i++ ) {
      if ( keys[i] != 0 ) {
	keycode = keymap[0][i].keycode & 0x2a000000;
	if ( keycode != 0 ) onekey( keycode, fdd );
      }
    }
    
    close( fdd );
    fdd = -1;
  }

  if (mousemode !=0 && xmode != 1){
    if(st == 1){
      mouseoff(prev_y,prev_x);
    }
  
  }

}


/*
  ˡɽ
*/
void	usage( int status )
{
  if ( status != 0 ) {
    printf( "-h for more information.\n" );
  }else{
    printf( "Usage: %s [OPTION]\n", program_name );
    
    printf(
	   "-p : serial port device name (defalt: /dev/ttyS0)\n"
	   "-k : vnc kbdsim module use\n"
	   "-i : debug info print\n"
	   "-d : direct mode\n"
	   "-b : background daemon\n"
	   "-v : print version\n"
	   "-h : display this message\n"
	   "-m : key mapping file name\n"
	   "-M : mouse mode\n"
           "-x : console/X mouse mode\n"
	   "     default: /opt/QtPalmtop/etc/ser_code_ksim.ini (or ser_code_direct.ini)\n"
	   );
  }
  
  exit( status );
}


/*
  ץࡦСɽ
*/
void	printversion( void )
{
  printf( "%s version %s\n", program_name, VERSION );
  exit( 0 );
}


/*
  ޥåסơ֥ɽ
*/
void	printinfo( void )
{
  int	i,j;
  
  for ( j = 0; j < 127; j++ ) {
    for ( i = 0; i < 15; i++ ) {
      if ( keymap[i][j].keycode != 0 ) {
	printf( "%02x: %02x : [%s], (%08lx)\n",
		i, 
		j, 
		keymap[i][j].keyname, 
		keymap[i][j].keycode
		);
      }
    }
  }
  
  for ( i = 0; i < 128; i++ ) {
    if ( metas[i] != 0 ) {
      printf( "metas[%02x] = %02x\n", i, metas[i] );
    }
  }
}


/*
	ե˴ŤơϽԤᥤ
*/
int	main( int argc, char **argv )
{
  unsigned char	rawcode;
  unsigned char	lastrawcode = 0;	// Ǹ˥ܡɤƤ
  
  int		daemonflag = 0;
  int		pressflag;
  int		meta;
  int		scancode;
  long		keycode;
  char		*keyname;
  int		mousecode = 0;
  int		mouse_x = 0;
  int		mouse_y = 0;
  int		mouse_x_dif = 0;
  int		mouse_y_dif = 0;
  int		mouse_x_sgn = 0;
  int		mouse_y_sgn = 0;
  int		mouse_x_of = 0;
  int		mouse_y_of = 0;
  int		mouse_mb = 0;
  int		mouse_lb = 0;
  int		mouse_rb = 0;
  int		raw_x; 
  int		raw_y; 
  
  // λˡץѥϥɥĤ뤿
  if ( atexit( close_all ) != 0 ) {
    usage( 2 );
  }
  
  // ư줿ʬȤ̾¸
  program_name = basename( argv[0] );
  
  
  // ϤơƼѿꤹ
  {
    int	i;
    int	j;
    int	k;
    
    for( i = 1; i < argc; i++ ) {
      if ( argv[i][0] != '-' ) {
	usage( 1 );
      }else{
	k = i;
	for( j = 1; j < strlen(argv[k]); j++ ) {
	  switch( argv[k][j] ) {
	  case 'p':
	    if ( ++i < argc ) serialdevice = argv[i]; else usage(1);
	    break;
	  case 'm':
	    if ( ++i < argc ) keymapfile = argv[i]; else usage(1);
	    break;
	  case 'k': vncflag = 1; break;
	  case 'i': infoflag = 1;	break;
	  case 'd': directflag = 1; break;
	  case 'b': daemonflag = 1; break;
	  case 'M': mousemode = 1; break;
	  case 'x': xmode = 1; break;
	  case 'v': printversion();
	  case 'h': usage( 0 );
	  default:  usage( 1 );
	  }
	}
      }
    }
  }

  // direct⡼ɤ̾⡼ɤǥǥեȤեѹ
  if ( keymapfile == NULL ) {
    keymapfile = (directflag) ? keymapfile_d : keymapfile_n;
  }

  // -i ե饰åȤƤ硢ɽ롣
  if ( infoflag ) {
    printf( "port = [%s]\nmap  = [%s]\n", serialdevice, keymapfile );
  }
  
  // եɤ߹ࡣ
  if ( read_key_mapping( keymapfile ) !=  RETOUR_OK ) {
    printf( "Key map file is not found\n" );
    //		exit( 1 );
  }
  
  // -i ե饰åȤƤ硢ɽ
  if ( infoflag ) {
    printinfo();
  }
  
  // kill ˥ץѥݡȤĤ뤿᥷ʥϥɥ
  signal( SIGINT,  &sig_handler );
  signal( SIGTERM, &sig_handler );
  signal( SIGHUP,  &sig_handler );
  
  // vncflag ξ֤ˤꡢ/dev/kbdsim /dev/sharp_kbdctl Ȥ
  {
    char * kbddrv = (vncflag ? VNCKBDDRV : SHARPKBDDRV);
    if ( (fdd = open( kbddrv, O_RDWR)) < 0 ) {
      printf( "Error opening %s device !\n", kbddrv );
      exit( 1 );
    }
  }

  //mouse
  loadConfig();
  //
  
  
  // ꥢݡȤΥץ
  if ( init_ser( serialdevice ) != RETOUR_OK ) {
    exit( 1 );
  }
  
  if ( infoflag ) {
    printf( "fds = %d, fdd = %d\n", fds, fdd );
  }
  
  // daemon
  if ( daemonflag ) {
    infoflag = 0;
    daemon( 0, 0 );
  }
  
  for(;;) {
    
    // 1ʸ Ͻ
    rawcode = get_1_char();
    
    if (mousemode ==0 ){
      if ( infoflag ) {
	printf( "raw data = %02x\n", (int)rawcode );
      }
      
      // 󥳡Τ߼Ф
      scancode = rawcode & 0x7f;
      
      // Ǿ̥ӥåȤ 1 ʤ饭Υ줿, 0 ʤ饭줿
      keys[ scancode ] = pressflag = (rawcode & (1 << 7)) ? 0 : 1;
      
      // ƤΥΥ줿ν
      // ( keys[] 򸫤ơ٤Ƥβɬפ³ʤΤǳ䰦)
      if ( (! pressflag) && (rawcode == lastrawcode) ) {
	metaflags   = 0;		// ƤΥ᥿Υ
	lastkeycode = -1;		// ԡȤ
      }
      lastrawcode = rawcode;
      
      if ( directflag ) {
	// 쥯ȥ⡼ɤǤϡnormal key Υɲ1byte򤽤ΤޤѤ롣
	keycode = keymap[0][scancode].keycode;
	keyname = keymap[0][scancode].keyname;
      }else{
	// Shift, Ctrl, Fn, Cmd ʤɤ metaˤĤե饰ꤹ롣
	
	// meta-key ʤСե饰ΩƤ
	meta = metas[ scancode ];
	
	if ( meta != 0 ) {
	  if ( pressflag ) {
	    metaflags |= meta;
	  }else{
	    metaflags &= ~meta;
	  }
	  continue;
	}
	
	// 겼ϡ̾省ν
	
	// ߤ Shift, Ctrl, Fn, Cmd б ͤ
	meta = (metaflags & 0x0f) | (metaflags >> 4);
	
	keycode = keymap[meta][scancode].keycode;
	keyname = keymap[meta][scancode].keyname;
      }
      
      // keycode  0 ξϡб륭̵Τǲ⤷ʤ
      
      if ( keycode == 0 ) {
	continue;
      }
      
      if ( directflag ) {
	if ( pressflag ) {
	  keycode &= 0x010000ff;
	}else{
	  keycode &= 0x020000ff;
	}
	onekey( keycode, fdd );
      }else{
	if ( pressflag ) {
	  onekey( keycode, fdd );
	  
	  lastkeycode = keycode;		// ԡȤΤ
	  keyrepeat   = keyrep_first;
	}else{
	  lastkeycode = -1;		// ԡȤߤΤ
	}
      }
    }else {
      switch (mousecode){
      case 0:
	mouse_y_of = (rawcode >> 7) & 0x01;
	mouse_x_of = (rawcode >> 6) & 0x01;
	mouse_y_sgn = (rawcode >> 5) & 0x01;
	mouse_x_sgn = (rawcode >> 4) & 0x01;
	mouse_mb = (rawcode >> 2) & 0x01;
	mouse_rb = (rawcode >> 1) & 0x01;
	mouse_lb = (rawcode >> 0) & 0x01;
	mousecode = 1;
	break;
      case 1:
	mouse_x_dif = rawcode;
	mousecode = 2;
	break;
      case 2:
	mouse_y_dif = rawcode;
	
	mouse_x = mouse_x + mouse_x_dif;
	mouse_y = mouse_y + mouse_y_dif;
	if (mouse_x_sgn)
	  mouse_x -= 256;
	if (mouse_y_sgn)
	  mouse_y -= 256;
	if(mouse_x > 639)
	  mouse_x = 640;
	if(mouse_y > 479)
	  mouse_y = 479;
	if(mouse_x < 0)
	  mouse_x = 0;
	if(mouse_y < 0)
	  mouse_y = 0;
	
	
	printf(" %03d %03d %03d ",mouse_x,mouse_y,mouse_lb);
	fflush(stdout);
	raw_y = (mouse_y * qcal.s * qcal.e - qcal.c * qcal.e) /(qcal.e * qcal.a);
	raw_x = (mouse_x * qcal.s * qcal.a - qcal.a * qcal.f) /(qcal.e * qcal.a);
	fp=fopen("/dev/ts", "w");
	if(xmode ==0){
	  fprintf(fp, "%d %d %d 0\n",mouse_lb*500,raw_y,raw_x);
	  printf("%03d %03d %03d 0\r",mouse_lb*500,raw_y,raw_x);
	}else{
	  fprintf(fp, "%d %d %d 0\n",(mouse_lb+1)*500,raw_x,480-raw_y);
	  printf("%03d %03d %03d 0\r",(mouse_lb+1)*500,raw_x,480-raw_y);
	}
	fclose(fp);
	if(xmode ==0){
	  if(st == 1){
            mouseoff(prev_y,prev_x);
          }
	  grabrect1(mouse_y,mouse_x);
	  st = 1;
          mouseon(mouse_y,mouse_x);
	  grabrect2(mouse_y,mouse_x);
	  prev_x =mouse_x;
	  prev_y =mouse_y;
	  
	}
	
	mousecode = 0;
	break;
      default:
	break;
      }
    }
  }
  exit( 0 );		// ɬפʥݡȤΥϡatexit 롼ǹԤ
}


/*
  VNCɥ饤ѤΥϽ
*/
void	onekey_v( int c, int fd, int pressflag )
{
  unsigned char buf[2];
  
  buf[0] = c; buf[1] = pressflag;
  write( fd, buf, 2 );
}


/*
  Sharp KeyCtrlɥ饤ѤΥϽ
*/
void	onekey_s( long c, int fd )
{
  ioctl( fd, SHARP_KBDCTL_SENDKEY, c );
}

/*
  Ͻ()
*/
int	onekey( long c, int fdd )
{
  if ( vncflag ) {
    if ( c & 0x10000000 ) onekey_v( (c >> 16) & 0xff, fdd, 1 );
    if ( c & 0x04000000 ) onekey_v( (c >> 8)  & 0xff, fdd, 1 );
    if ( c & 0x01000000 ) onekey_v( (c)       & 0xff, fdd, 1 );
    
    if ( c & 0x02000000 ) onekey_v( (c)       & 0xff, fdd, 0 );
    if ( c & 0x08000000 ) onekey_v( (c >> 8)  & 0xff, fdd, 0 );
    if ( c & 0x20000000 ) onekey_v( (c >> 16) & 0xff, fdd, 0 );
  }else{
    onekey_s( c, fdd );
  }
}

